@@ -1,6 +1,6 @@
 ;;  -*-  indent-tabs-mode:nil; coding: utf-8 -*-
-;;  Copyright (C) 2013,2014,2015,2016,2017,2018,2019,2020
-;;      "Mu Lei" known as "NalaGinrut" <NalaGinrut@gmail.com>
+;;  Copyright (C) 2013-2025
+;;      "Mu Lei" known as "NalaGinrut" <mulei@gnu.org>
 ;;  Artanis is free software: you can redistribute it and/or modify
 ;;  it under the terms of the GNU General Public License and GNU
 ;;  Lesser General Public License published by the Free Software
@@ -18,7 +18,6 @@
 ;;  If not, see <http://www.gnu.org/licenses/>.
 
 (define-module (artanis utils)
-  #:use-module (artanis security nss)
   #:use-module (artanis tpl sxml)
   #:use-module (artanis config)
   #:use-module (artanis irregex)
@@ -35,6 +34,8 @@
   #:use-module (ice-9 receive)
   #:use-module (ice-9 q)
   #:use-module (ice-9 control)
+  #:use-module (ice-9 threads)
+  #:use-module (ice-9 control)
   #:use-module (srfi srfi-1)
   #:use-module (srfi srfi-19)
   #:use-module (web http)
@@ -91,7 +92,7 @@
             colorize-string WARN-TEXT ERROR-TEXT REASON-TEXT
             NOTIFY-TEXT STATUS-TEXT get-trigger get-family get-addr request-path
             response-keep-alive? request-keep-alive? file->bytevector
-            procedure-name->string proper-toplevel gen-content-length
+            procedure-name->string gen-content-length
             make-file-sender file-sender? file-sender-size file-sender-thunk
             get-string-all-with-detected-charset make-unstop-exception-handler
             artanis-log exception-from-client exception-from-server render-sys-page
@@ -100,7 +101,9 @@
             is-guile-compatible-server-core? positive-integer? negative-integer?
             io-exception:peer-is-shutdown? io-exception:out-of-memory?
             out-of-system-resources? allow-long-live-connection?
-            free-JS-announcement generate-rule-uid gen-cache-file)
+            free-JS-announcement gen-cache-file current-route-cache
+            gnu-locale->http-lang-tag http-lang-tag->gnu-locale
+            sxml-query sxml-replace!)
   #:re-export (the-environment
                utf8->string
                bytevector?
@@ -126,16 +129,7 @@
                request-headers
                request-method
                request-content-length
-               request-port
-
-               (nss:md5 . string->md5)
-               (nss:sha-1 . string->sha-1)
-               (nss:sha-224 . string->sha-224)
-               (nss:sha-256 . string->sha-256)
-               (nss:sha-384 . string->sha-384)
-               (nss:sha-512 . string->sha-512)
-               nss:base64-encode
-               nss:base64-decode))
+               request-port))
 
 (define parse-date (@@ (web http) parse-date))
 (define write-date (@@ (web http) write-date))
@@ -304,7 +298,10 @@
   ;; TODO: support AF_INET6 in the future
   (if (port-filename (request-port req))
       ;; Valid socket port
-      (inet-ntop AF_INET (sockaddr:addr (getpeername (request-port req))))
+      (catch #t
+        (lambda ()
+          (inet-ntop AF_INET (sockaddr:addr (getpeername (request-port req)))))
+        (lambda _ "Unknown IP"))
       "localtest")) ; fake hostname for testing
 
 (define-syntax-rule (remote-info req)
@@ -772,9 +769,15 @@
                    "Fatal! Can't be here `~a'!" lst))))
   (define (-> x)
     (string-trim-both x (lambda (c) (member c '(#\sp #\: #\return)))))
+  (define (fix x)
+    (cond
+     ((string? x) x)
+     ((bytevector? x) (utf8->string x))
+     (else (throw 'artanis-err 500 generate-kv-from-post-qstr
+                  "Fatal! Can't be here `~a'!" x))))
   (map (lambda (x)
          (%convert (map -> (string-split (cook x) #\=))))
-       (string-split (utf8->string body) #\&)))
+       (string-split (fix body) #\&)))
 
 ;; NOTE: We accept warnings, which means if warnings occurred, it'll be 200(OK) anyway, but
 ;;       Artanis will throw warnings in the server-side log.
@@ -866,11 +869,16 @@
   (define (write-header port)
     (format port ";; Do not touch anything!!!~%")
     (format port ";; All things here should be automatically handled properly!!!~%"))
-  (define route-cache (string-append (current-toplevel) "/tmp/cache/route.cache"))
+  (define cache-dir (format #f "~a/cache" (current-tmp)))
+  (define route-cache (format #f "~a/cache/route.cache" (current-tmp)))
   (when (not (file-exists? route-cache))
     (format (artanis-current-output) "Regenerating route cache ...~%")
-    (call-with-output-file route-cache
-      (lambda (port) (write-header port) (write '() port))))
+    (when (not (file-exists? cache-dir))
+      (mkdir cache-dir))
+    (let ((port (open-file route-cache "w")))
+      (write-header port)
+      (write '() port)
+      (close port)))
   (when (and url meta)
     (let ((rl (call-with-input-file route-cache read)))
       (delete-file route-cache)
@@ -890,10 +898,13 @@
                ))
           (flock port LOCK_UN))))))
 
+(define (current-route-cache)
+  (format #f "~a/.route" (current-tmp)))
+
 (define (dump-route-from-cache)
   (define toplevel (current-toplevel))
-  (define route-cache (string-append toplevel "/tmp/cache/route.cache"))
-  (define route (string-append toplevel "/.route"))
+  (define route-cache (format #f "~a/cache/route.cache" (current-tmp)))
+  (define route (current-route-cache))
   (define (load-customized-router)
     (let ((croute (string-append toplevel "conf/route")))
       (cond
@@ -1078,23 +1089,26 @@
   (if (not (provided? 'posix))
       (lambda (thunk handler-thunk) (thunk))
       (lambda (thunk handler-thunk)
-        (let ((handler #f))
+        (let ((handler (make-parameter #f)))
           (catch 'interrupt
             (lambda ()
               (dynamic-wind
-                (lambda ()
-                  (set! handler
-                        (sigaction SIGINT (lambda (sig)
-                                            (run-when-sigint-hook)
-                                            (throw 'interrupt)))))
-                thunk
-                (lambda ()
-                  (if handler
-                      ;; restore Scheme handler, SIG_IGN or SIG_DFL.
-                      (sigaction SIGINT (car handler) (cdr handler))
-                      ;; restore original C handler.
-                      (sigaction SIGINT #f)))))
-            (lambda (k . _) (handler-thunk)))))))
+                  (lambda ()
+                    (handler (sigaction SIGINT))
+                    (sigaction SIGINT (lambda (sigtmp)
+                                        (run-when-sigint-hook)
+                                        (throw 'interrupt))))
+                  thunk
+                  (lambda ()
+                    (if (handler)
+                        ;; restore Scheme handler, SIG_IGN or SIG_DFL.
+                        (sigaction SIGINT (car (handler)) (cdr (handler)))
+                        ;; restore original C handler.
+                        (sigaction SIGINT #f)))))
+            (lambda (k . _)
+              (handler-thunk)
+              ((car (handler)) SIGINT)
+              (primitive-exit 0)))))))
 
 (define-syntax-rule (define-box-type name)
   (define-record-type name (fields treasure)))
@@ -1166,6 +1180,13 @@
       (lambda (v e)
         (or (eq? e 'ANY)
             (eq? (detect-type-name v) e)
+            (match e
+              (('or types ...)
+               (any (lambda (t)
+                      (or (equal? t v)
+                          (eq? t (detect-type-name v))))
+                    types))
+              (else #f))
             (throw 'artanis-err 500 check-function-types
                    "`Return value ~a(~a) is expected to be type `~a'"
                    v (detect-type-name v) e)))
@@ -1299,9 +1320,6 @@
 (define (procedure-name->string proc)
   (symbol->string (procedure-name proc)))
 
-(define-syntax-rule (proper-toplevel)
-  (or (current-toplevel) ""))
-
 (define-record-type file-sender
   (fields size thunk))
 
@@ -1343,21 +1361,22 @@
 ;; ENHANCE: use colored output
 (define* (artanis-log blame-who? status mime #:key (port (current-error-port))
                       (request #f))
-  (case blame-who?
-    ((client)
-     (when (not request)
-       (error "artanis-log: Fatal bug! Request shouldn't be #f here!~%"))
-     (let* ((uri (request-uri request))
-            (path (uri-path uri))
-            (qstr (uri-query uri))
-            (method (request-method request)))
-       (format port "[Remote] ~a @ ~a~%" (remote-info request) (local-time-stamp))
-       (format port "[Request] method: ~a, path: ~a, query: ~a~%" method path qstr)
-       (format port "[Response] status: ~a, MIME: ~a~%~%" status mime)))
-    ((server)
-     (format port "[Server] ~a @ ~a~%" (get-conf '(host addr)) (local-time-stamp))
-     (format port "[Response] status: ~a, MIME: ~a~%~%" status mime))
-    (else (error "artanis-log: Fatal BUG here!"))))
+  (monitor
+   (case blame-who?
+     ((client)
+      (when (not request)
+        (error "artanis-log: Fatal bug! Request shouldn't be #f here!~%"))
+      (let* ((uri (request-uri request))
+             (path (uri-path uri))
+             (qstr (uri-query uri))
+             (method (request-method request)))
+        (format port "[Remote] ~a @ ~a~%" (remote-info request) (local-time-stamp))
+        (format port "[Request] method: ~a, path: ~a, query: ~a~%" method path qstr)
+        (format port "[Response] status: ~a, MIME: ~a~%~%" status mime)))
+     ((server)
+      (format port "[Server] ~a @ ~a~%" (get-conf '(host addr)) (local-time-stamp))
+      (format port "[Response] status: ~a, MIME: ~a~%~%" status mime))
+     (else (error "artanis-log: Fatal BUG here!")))))
 
 (define *guile-compatible-server-core*
   '(guile fibers))
@@ -1600,10 +1619,6 @@
     </script>
 ")
 
-;; FIXME: MD5 may not be the best choice
-(define-syntax-rule (generate-rule-uid rule)
-  (string->md5 rule))
-
 (define (gen-cache-file path)
   (define-syntax-rule (-> str)
     (string-trim-both
@@ -1611,5 +1626,41 @@
      (lambda (x) (memv x '(#\sp #\_)))))
   (let ((p (-> path)))
     (if (string-null? p)
-        (format #f "~a/tmp/cache/index.html" (current-toplevel))
-        (format #f "~a/tmp/cache/~a.html" (current-toplevel) (-> path)))))
+        (format #f "~a/cache/index.html" (current-tmp))
+        (format #f "~a/cache/~a.html" (current-tmp) (-> path)))))
+
+(define-syntax-rule (sxml-query pattern alst)
+  (let/ec return
+    (fold
+     (lambda (x p)
+       (or (assoc-ref p x)
+           (return #f)))
+     alst
+     (string-split pattern #\/))))
+
+(define (sxml-replace! alst key val)
+  (define (list-deep-copy lst)
+    (map (lambda (x) (if (list? x) (list-deep-copy x) x)) lst))
+  (let ((new-lst (list-deep-copy alst))
+        (v (assoc-ref alst key)))
+    (cond
+     ((null? v)
+      (assoc-set! new-lst key val))
+     (else
+      (assoc-set! new-lst key (cons val v))))))
+
+(define *http-lang-tag-re* (string->irregex "([a-z]+)-([A-Z])+"))
+(define (http-lang-tag->gnu-locale lang)
+  ;; replace - with _
+  (if (irregex-match *http-lang-tag-re* lang)
+      (irregex-replace/all "-" lang "_")
+      (throw 'artanis-err 500 http-lang-tag->gnu-locale
+             "Invalid HTTP lang tag format: ~a" lang)))
+
+(define *gnu-locale-re* (string->irregex "([a-z]+)_([A-Z])+"))
+(define (gnu-locale->http-lang-tag lang)
+  ;; replace _ with -
+  (if (irregex-match *gnu-locale-re* lang)
+      (irregex-replace/all "_" lang "-")
+      (throw 'artanis-err 500 gnu-locale->http-lang-tag
+             "Invalid GNU locale format: ~a" lang)))
